import math
from typing import List
from BuildingElement import BuildingElement

import NemAll_Python_BasisElements as AllplanBasisElements
import NemAll_Python_BaseElements as AllplanBaseElements
import NemAll_Python_Geometry as AllplanGeo
import NemAll_Python_IFW_Input as AllplanIFW

from BuildingElementPaletteService import BuildingElementPaletteService
from BuildingElementService import BuildingElementService

def check_allplan_version(build_ele, version):
    """
    Check the current Allplan version

    Args:
        build_ele: the building element.
        version:   the current Allplan version

    Returns:
        True/False if version is supported by this script
    """

    # Delete unused arguments
    del build_ele
    del version

    # Support all versions
    return True

def create_element(build_ele, doc):
    """use to create preview"""
    return ([], [])

def create_interactor(coord_input: AllplanIFW.CoordinateInput, pyp_path: str, str_table_service):
    """
    Start interactor
    Args:
        coord_input: coordinate input
        pyp_path: path of the pyp file
        str_table_service: string table service
    """

    return StripFoundationClass(coord_input, pyp_path, str_table_service)

def move_pnt_to_perpendicular(ref_pnt, distance, angle):
    """move reference point by distance that perpendicular with given angle"""
    new_x = ref_pnt.X + distance*math.cos(angle + math.pi/2)
    new_y = ref_pnt.Y + distance*math.sin(angle + math.pi/2)
    return AllplanGeo.Point3D(new_x, new_y, ref_pnt.Z)

class StripFoundationClass():
    def __init__(self, coord_input: AllplanIFW.CoordinateInput, pyp_path: str, str_table_service):

        self.coord_input = coord_input
        self.str_table_service = str_table_service
        self.build_ele_service = BuildingElementService()

        #show the palette
        self.is_success = self.show_palette(pyp_path)
        if not self.is_success:
            return

        # Selected points
        self.selected_pnts = []

        # Show input message
        self.reset()

    def reset(self):
        """reset selection for next foundation"""
        self.selected_pnts.clear()
        self.coord_input.InitFirstPointInput(AllplanIFW.InputStringConvert("Select the first point to create foundation"))

    def show_palette(self, pyp_path) -> bool:
        """read data from pyp file and show the palette
           Return: bool for success
        """
        result, self.build_ele_script, self.build_ele_list, self.control_props_list,    \
            self.build_ele_composite, part_name, self.file_name = \
            self.build_ele_service.read_data_from_pyp(pyp_path + "\\StripFoundation.pyp", self.str_table_service.str_table, False, 
                                                      self.str_table_service.material_str_table)

        if not result:
            return False

        self.palette_service = BuildingElementPaletteService(self.build_ele_list, self.build_ele_composite,
                                                             self.build_ele_script,
                                                             self.control_props_list, self.file_name)

        self.palette_service.show_palette(part_name)

        return True

    def process_mouse_msg(self,
                          mouse_msg: int,
                          pnt      : AllplanGeo.Point2D,
                          msg_info : AllplanIFW.AddMsgInfo) -> bool:
        """
        Handles the process mouse message event

        Args:
            mouse_msg: the mouse message.
            pnt      : the input point.
            msg_info : additional message info.

        Returns:
            True/False for success.
        """

        # Get input point
        if self.selected_pnts:
            input_pnt = self.coord_input.GetInputPoint(mouse_msg, pnt, msg_info, self.selected_pnts[-1], True).GetPoint()
        else:
            input_pnt = self.coord_input.GetInputPoint(mouse_msg, pnt, msg_info).GetPoint()

        if self.coord_input.IsMouseMove(mouse_msg):
            return True

        # Save input point to selected points
        self.selected_pnts.append(input_pnt)

        # Create model and reset for next use
        if len(self.selected_pnts) == 2:
            self.create_models(self.build_ele_list[0], self.selected_pnts[0], self.selected_pnts[1])
            self.reset()

        # Message for second point
        if len(self.selected_pnts) == 1:
            self.coord_input.InitNextPointInput(AllplanIFW.InputStringConvert("Select the second point to create foundation"))

        return True

    def on_cancel_function(self) -> bool:
        """
        Cancel the input function

        Returns:
            True/False for success.
        """

        if self.is_success:
            self.palette_service.close_palette()
        return True

    def on_mouse_leave(self):
        """
        Handles the mouse leave event
        """
        pass

    def on_preview_draw(self):
        """
        Handles the preview draw event
        """
        pass

    def modify_element_property(self,
                                page : str,
                                name : str,
                                value):
        """
        Modify property of element

        Args:
            page:   the page of the property
            name:   the name of the property.
            value:  new value for property.
        """
        update_palette = self.palette_service.modify_element_property(page, name, value)

        if update_palette:
            self.palette_service.update_palette(-1, False)

    def create_models(self, build_ele: BuildingElement, first_pnt: AllplanGeo.Point3D, second_pnt: AllplanGeo.Point3De):
        """create models contain foundation and wall"""

        #common parameters
        length = first_pnt.GetDistance(second_pnt)
        angle = AllplanGeo.CalcAngle(AllplanGeo.Point2D(first_pnt), AllplanGeo.Point2D(second_pnt))

        #list of geometries
        elements = []
        elements += self.create_foundation(build_ele, length, angle)
        elements += self.create_wall(build_ele, length, angle)

        #create elements
        AllplanBaseElements.CreateElements(self.coord_input.GetInputViewDocument(), AllplanGeo.Matrix3D(),
                                                elements, [], None)

    def create_foundation(self, build_ele: BuildingElement, length: float, angle: AllplanGeo.Angle) -> List[AllplanBasisElements.ModelElement3D]:
        """
        Create foundation model

        Args:
            build_ele: the building element

        Return: List of ModelElement3D
        """
        common_props = self.get_foundation_common_props(self.build_ele_list[0])
        geometry = AllplanGeo.Polyhedron3D.CreateCuboid(length, #length
                                                        build_ele.FounWidth.value,  #width
                                                        build_ele.FounHeight.value) #height

        #transform geometry
        geometry = self.transform_geometry(geometry, self.selected_pnts[0], angle)

        return [AllplanBasisElements.ModelElement3D(common_props, geometry)]

    def create_wall(self, build_ele: BuildingElement, length: float, angle: AllplanGeo.Angle) -> List[AllplanBasisElements.ModelElement3D]:
        """
        Create wall model

        Args:
            build_ele: the building element

        Return: List of ModelElement3D
        """
        common_props = self.get_wall_common_props(self.build_ele_list[0])
        geometry = AllplanGeo.Polyhedron3D.CreateCuboid(length, #length
                                                        build_ele.WallWidth.value,  #width
                                                        build_ele.WallHeight.value) #height

        #transform geometry
        ref_pnt = move_pnt_to_perpendicular(self.selected_pnts[0], build_ele.FounWidth.value/2 - build_ele.WallWidth.value/2, angle) \
                                            + AllplanGeo.Point3D(0,0,build_ele.FounHeight.value)
        geometry = self.transform_geometry(geometry, ref_pnt, angle)

        return [AllplanBasisElements.ModelElement3D(
        common_props, geometry)]

    def get_foundation_common_props(self, build_ele: BuildingElement) -> AllplanBaseElements.CommonProperties:
        """
        Get foundation common properties

        Return: AllplanBaseElements.CommonProperties
        """

        common_prop = AllplanBaseElements.CommonProperties()

        common_prop.Color   = build_ele.FounColor.value
        common_prop.Pen     = build_ele.FounPen.value
        common_prop.Stroke  = build_ele.FounStroke.value
        common_prop.Layer   = build_ele.FounLayer.value

        #use global properties
        if build_ele.IsFounUseGlobalProp.value:
            common_prop.GetGlobalProperties()

        return common_prop

    def get_wall_common_props(self, build_ele: BuildingElement) -> AllplanBaseElements.CommonProperties:
        """
        Get wall common properties

        Return: AllplanBaseElements.CommonProperties
        """

        common_prop = AllplanBaseElements.CommonProperties()

        common_prop.Color   = build_ele.WallColor.value
        common_prop.Pen     = build_ele.WallPen.value
        common_prop.Stroke  = build_ele.WallStroke.value
        common_prop.Layer   = build_ele.WallLayer.value

        #use global properties
        if build_ele.IsFounUseGlobalProp.value:
            common_prop.GetGlobalProperties()

        return common_prop

    def transform_geometry(self, geometry: AllplanGeo.Polyhedron3D, ref_pnt: AllplanGeo.Point3D, angle: AllplanGeo.Angle) -> AllplanGeo.Polyhedron3D:
        """move geometry to reference point and rotate base on selected points"""

        #move to reference point
        geometry = AllplanGeo.Move(geometry, AllplanGeo.Vector3D(ref_pnt))

        #rotate
        axis3d = AllplanGeo.Axis3D(ref_pnt, AllplanGeo.Vector3D(0,0,1))
        geometry = AllplanGeo.Rotate(geometry, axis3d, angle)

        return geometry
